package cz.fidentis.analyst.mesh.KdTree;

import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshPoint;

import java.util.*;

public class KdTreeImpl {
    private  KdNode root;
    private List<MeshFacet> facets;


    private Map<MeshPoint, List<IndexOfPoint>> samePoints = new HashMap<>();

    private final static int X_AXIS = 0;
    private final static int Y_AXIS = 1;
    private final static int Z_AXIS = 2;


    private static class IndexOfPoint {
        private Integer indexInFacet;
        private MeshFacet facet;

        public IndexOfPoint( int indexInFacet,  MeshFacet facet) {
            this.indexInFacet = indexInFacet;

            this.facet = facet;
        }

        public Integer getIndexInFacet() {
            return indexInFacet;
        }

        public MeshFacet getFacet() {
            return facet;
        }

    }

    private static class Triple {
        private MeshFacet facet;
        private MeshPoint point;
        private Integer index;

        public Triple(MeshFacet facet, MeshPoint point, int index) {
            this.facet = facet;
            this.point = point;
            this.index = index;
        }

        public MeshFacet getFacet() {
            return facet;
        }

        public MeshPoint getPoint() {
            return point;
        }

        public Integer getIndex() {
            return index;
        }
    }



    private final Comparator<Integer> xIndexComparatorF = (i1, i2) -> {
        if (Objects.requireNonNull(findPointFacetIndex(i1)).getPoint().getPosition().x < Objects.requireNonNull(findPointFacetIndex(i2)).getPoint().getPosition().x)
            return -1;
        if (Objects.requireNonNull(findPointFacetIndex(i1)).getPoint().getPosition().x > Objects.requireNonNull(findPointFacetIndex(i2)).getPoint().getPosition().x)
            return 1;

        createSamePoints(i1,i2);
        return 0;
    };

    private final Comparator<Integer> yIndexComparatorF = Comparator.comparingDouble(i -> Objects.requireNonNull(findPointFacetIndex(i)).getPoint().getPosition().y);

    private final Comparator<Integer> zIndexComparatorF = Comparator.comparingDouble(i -> Objects.requireNonNull(findPointFacetIndex(i)).getPoint().getPosition().z);


    public  KdTreeImpl(List<MeshFacet> facets) {
        if(facets == null ||  facets.isEmpty() || facets.get(0).getVertices().isEmpty() ){
            this.root = null;
            this.facets = null;
            return;
        }

        int countOfVertices = 0;
        for (MeshFacet facet : facets) {
            countOfVertices += facet.getVertices().size();
        }



        this.facets = facets;

        List<Integer> sortedByX = sortPointsF( X_AXIS);
        List<Integer> sortedByY = sortPointsF( Y_AXIS);
        List<Integer> sortedByZ = sortPointsF( Z_AXIS);

        List<Integer> uniqueSortedX = new LinkedList<>();
        List<Integer> uniqueSortedY = new LinkedList<>();
        List<Integer> uniqueSortedZ = new LinkedList<>();


        if (!samePoints.isEmpty()) {
            for (int i = 0; i < countOfVertices; i++) {
                if (samePoints.containsKey(Objects.requireNonNull(findPointFacetIndex(sortedByX.get(i))).getPoint())) {
                    Objects.requireNonNull(findPointFacetIndex(sortedByX.get(i)));
                    uniqueSortedX.add(sortedByX.get(i));
                }
                if (samePoints.containsKey(Objects.requireNonNull(findPointFacetIndex(sortedByY.get(i))).getPoint())) {
                    Objects.requireNonNull(findPointFacetIndex(sortedByY.get(i)));
                    uniqueSortedY.add(sortedByY.get(i));
                }
                if (samePoints.containsKey(Objects.requireNonNull(findPointFacetIndex(sortedByZ.get(i))).getPoint())) {
                    Objects.requireNonNull(findPointFacetIndex(sortedByZ.get(i)));
                    uniqueSortedZ.add(sortedByZ.get(i));
                }
            }

            root = buildTreeF(null, uniqueSortedX, uniqueSortedY, uniqueSortedZ, 0);
        } else {
            root = buildTreeF(null, sortedByX, sortedByY, sortedByZ, 0);
        }
    }

    private Triple findPointFacetIndex(int index) {
        Triple result;
        int countOfCheckedVertices = 0;
        for (MeshFacet facet : facets) {
            if (countOfCheckedVertices + facet.getVertices().size() > index) {
                result = new Triple(facet, facet.getVertex(index - countOfCheckedVertices), index - countOfCheckedVertices);
                return result;
            }
            countOfCheckedVertices += facet.getVertices().size();
        }
        return null;
    }


    private boolean checkSamePoints(MeshPoint p1, MeshPoint p2){
        return p1.getPosition().x == p2.getPosition().x
                && p1.getPosition().y == p2.getPosition().y
                && p1.getPosition().z == p2.getPosition().z;
    }

    public void createSamePoints(Integer i1, Integer i2) {
        Triple firstP = findPointFacetIndex(i1);
        Triple secondP = findPointFacetIndex(i2);
        List<IndexOfPoint> allFacetsOfPoint = new LinkedList<>();

        if(checkSamePoints(Objects.requireNonNull(firstP).getPoint(), Objects.requireNonNull(secondP).getPoint())) {

            if(samePoints.containsKey(firstP.getPoint())) {
                allFacetsOfPoint.addAll(samePoints.get(firstP.getPoint()));
            } else {
                allFacetsOfPoint.add(new IndexOfPoint(firstP.getIndex(), firstP.getFacet()));
            }

            allFacetsOfPoint.add(new IndexOfPoint(secondP.getIndex(),secondP.getFacet()));
            samePoints.putIfAbsent(firstP.getPoint(), allFacetsOfPoint);
        }
    }



    private List<Integer> sortPointsF(int level){
        int countOfVertices = 0;
        for (MeshFacet facet : facets) {
            countOfVertices += facet.getVertices().size();
        }
        List<Integer> sortedListIndex = new ArrayList<>(countOfVertices);
        for(int i = 0; i < countOfVertices; i++){
            sortedListIndex.add(i);
        }

        if(level % 3 == 0){
            sortedListIndex.sort(xIndexComparatorF);
        }else if(level % 3 == 1){
            sortedListIndex.sort(yIndexComparatorF);
        }else if(level % 3 == 2){
            sortedListIndex.sort(zIndexComparatorF);
        }
        return sortedListIndex;

    }

    private boolean comparePointsOnLevel(MeshPoint p1, MeshPoint p2, int level){
        if(level % 3 == 0){
            return p1.getPosition().x <= p2.getPosition().x;
        }else if(level % 3 == 1){
            return p1.getPosition().y <= p2.getPosition().y;
        }else if(level % 3 == 2){
            return p1.getPosition().z <= p2.getPosition().z;
        }

        return false;
    }


    private void setFacetsOfNode(KdNode node, int midIndex) {
        for (int i = 0; i < samePoints.get(Objects.requireNonNull(findPointFacetIndex(midIndex)).getPoint()).size(); i++) {
            IndexOfPoint point = samePoints.get(Objects.requireNonNull(findPointFacetIndex(midIndex)).getPoint()).get(i);
            node.addFacet(point.getFacet(), point.getIndexInFacet());
        }
    }


    private KdNode buildTreeF(KdNode parent, List<Integer> byX, List<Integer> byY, List<Integer> byZ, int level) {
        KdNode node = null;
        int mid;
        int midIndex;

        if (byX.size() > 0 && byY.size() > 0 && byZ.size() > 0) {
            mid = (byX.size() / 2);

            List<Integer> leftX = new ArrayList<>(mid);
            List<Integer> leftY = new ArrayList<>(mid);
            List<Integer> leftZ = new ArrayList<>(mid);

            //split lists in half, set middle element as new KdNode to be returned
            //first list to be split based on level, rest so that they contain points in first list
            //but also keep the ordering by their axis

            if (level % 3 == 0) {
                midIndex = byX.get(mid);
                node = new KdNode(Objects.requireNonNull(findPointFacetIndex(midIndex)).getFacet(), Objects.requireNonNull(findPointFacetIndex(midIndex)).getIndex(), level, parent);

                if (samePoints.containsKey(Objects.requireNonNull(findPointFacetIndex(midIndex)).getPoint())) {
                   setFacetsOfNode(node, midIndex);
                }

                splitTree(mid, leftX, byX, leftY, byY, leftZ, byZ);

            } else if (level % 3 == 1) {
                midIndex = byY.get(mid);
                node = new KdNode(Objects.requireNonNull(findPointFacetIndex(midIndex)).getFacet(), Objects.requireNonNull(findPointFacetIndex(midIndex)).getIndex(), level, parent);

                if (samePoints.containsKey(Objects.requireNonNull(findPointFacetIndex(midIndex)).getPoint())) {
                    setFacetsOfNode(node, midIndex);
                }

                splitTree(mid, leftY, byY, leftX, byX, leftZ, byZ);

            } else {
                midIndex = byZ.get(mid);
                node = new KdNode(Objects.requireNonNull(findPointFacetIndex(midIndex)).getFacet(), Objects.requireNonNull(findPointFacetIndex(midIndex)).getIndex(), level, parent);

                if (samePoints.containsKey(Objects.requireNonNull(findPointFacetIndex(midIndex)).getPoint())) {
                    setFacetsOfNode(node, midIndex);
                }

                splitTree(mid, leftZ, byZ, leftY, byY, leftX, byX);

            }

            //removes current middle node from each all lists, in case there were duplicates
            byX.removeAll(Collections.singleton(midIndex));
            byY.removeAll(Collections.singleton(midIndex));
            byZ.removeAll(Collections.singleton(midIndex));
            leftX.removeAll(Collections.singleton(midIndex));
            leftY.removeAll(Collections.singleton(midIndex));
            leftZ.removeAll(Collections.singleton(midIndex));

            node.setLesser(buildTreeF(node, leftX, leftY, leftZ, level + 1));
            node.setGreater(buildTreeF(node, byX, byY, byZ, level + 1));
        }

        return node;
    }


    private void splitTree(int mid, List<Integer> leftMain, List<Integer> mainList, List<Integer> leftSecond,
                           List<Integer> secondList, List<Integer> leftThird, List<Integer> thirdList){
        for (int i = 0; i < mid; i++) {

            leftMain.add(mainList.get(0));

            mainList.remove(0);
        }

        splitArray(leftMain, leftSecond, secondList);
        splitArray(leftMain, leftThird, thirdList);
    }

    private void splitArray(List<Integer> leftMain, List<Integer> leftSecond, List<Integer> secondList){
        for(int i = secondList.size() - 1; i >= 0; i--){
            if(leftMain.contains(secondList.get(i))){
                leftSecond.add(0, secondList.get(i));
                secondList.remove(i);
            }
        }
    }
    public MeshPoint nearestNeighbour(MeshPoint p){
        KdNode nn = nearestNeighbourNode(p);

        if(nn == null){
            return null;
        }

        return nn.getId();
    }


    public boolean containPoint(MeshPoint p){
        MeshPoint found = nearestNeighbour(p);

        if(found == null){
            return false;
        }

        return found.equals(p);
    }

     public double distancePoints(MeshPoint point1, MeshPoint point2) {
        return Math.sqrt(Math.pow(point1.getPosition().x - point2.getPosition().x, 2)
                + Math.pow(point1.getPosition().y - point2.getPosition().y, 2)
                + Math.pow(point1.getPosition().z - point2.getPosition().z, 2));
    }

    private double minDistanceIntersection(MeshPoint searched, MeshPoint p2, int level){
        if(level % 3 == 0){
            return Math.abs(searched.getPosition().x - p2.getPosition().x);
        }else if(level % 3 == 1){
            return Math.abs(searched.getPosition().y - p2.getPosition().y);
        }else{
            return Math.abs(searched.getPosition().z - p2.getPosition().z);
        }
    }



    public KdNode nearestNeighbourNode(MeshPoint p) {
        if(p == null || this.root == null){
            return null;
        }

        double minDistance = Double.MAX_VALUE;
        double distOnAxis;
        KdNode near = root;
        KdNode searched = root;

        Queue<KdNode> queue = new LinkedList<>();

        while(searched != null){
            if(distancePoints(p, searched.getId()) < minDistance){
                near = searched;
                minDistance = distancePoints(p, searched.getId());
            }

            if(comparePointsOnLevel(p, searched.getId(), searched.getDepth())){
                searched = searched.getLesser();
            }else{
                searched = searched.getGreater();
            }
        }

        queue.add(root);

        //second search to find vertex that could be potentially closer than
        //nearest vertex already found
        while (!queue.isEmpty()) {

            //if min distance is 0, break cycle (nothing can be closer)
            if(minDistance == 0){
                break;
            }

            searched = queue.poll();

            if (distancePoints(p, searched.getId()) < minDistance) {
                near = searched;
                minDistance = distancePoints(p, searched.getId());
            }

            distOnAxis = minDistanceIntersection(searched.getId(), p, searched.getDepth());

            if (distOnAxis > minDistance) {
                if (comparePointsOnLevel(p, searched.getId(), searched.getDepth())) {
                    if (searched.getLesser() != null) {
                        queue.add(searched.getLesser());
                    }
                } else {
                    if (searched.getGreater() != null) {
                        queue.add(searched.getGreater());
                    }
                }
            } else {
                if (searched.getLesser() != null) {
                    queue.add(searched.getLesser());
                }
                if (searched.getGreater() != null) {
                    queue.add(searched.getGreater());
                }
            }
        }

        return near;

    }

    public Integer nearestNeighbourCountOfFacets(MeshPoint p){
        KdNode nn = nearestNeighbourNode(p);

        if(nn == null){
            return null;
        }

        return nn.getFacets().size();
    }

    /**
     * Find index of MeshPoint in MeshFacetImpl -> vertices
     * @param p MeshPoint
     * @return index of MeshPoint in MeshFacetImpl -> vertices
     */
    public Integer nearestNeighbourIndex(MeshPoint p){
        KdNode nn = nearestNeighbourNode(p);

        if(nn == null){
            return null;
        }

        return nn.getIndex();
    }

}
